"""Custom logger."""

import enum
import logging
import os
import sys

from typing import ClassVar


class TerminalFormat(enum.Enum):
    """Bash terminal format codes."""

    RED: ClassVar[str] = "\033[91m"
    YELLOW: ClassVar[str] = "\033[93m"

    BACKGROUND_RED: ClassVar[str] = "\33[41m"

    END: ClassVar[str] = "\033[0m"


class ColoredFormatter(logging.Formatter):
    """Colorizes the logging level when it is used in the format string."""

    MAPPING: ClassVar[dict[str, TerminalFormat]] = {
        "WARNING": TerminalFormat.YELLOW,
        "ERROR": TerminalFormat.RED,
        "CRITICAL": TerminalFormat.BACKGROUND_RED,
    }

    def format(self, record: logging.LogRecord) -> str:
        """Format a log record

        :param record: The log record to format

        :returns: The formatted record
        """

        pre_formatted = super().format(record)

        if not sys.stdout.isatty():
            return pre_formatted

        color = ColoredFormatter.MAPPING.get(record.levelname)

        if color is None:
            return pre_formatted

        return f"{color.value}{pre_formatted}{TerminalFormat.END.value}"


def _configure(logger: logging.Logger, with_decorators: bool = True) -> None:
    """Configure the logger.

    :param logger: The logger to configure
    :param filename: The file to write the log contents out to
    :param with_decorators: If False only the message will be printed (no timestamp or level name)
    """

    # Base logging level. Handlers can set their own ones
    logger.setLevel(logging.DEBUG)
    # Do not pass logging messages to the handlers of ancestor loggers
    logger.propagate = False

    if with_decorators:
        format_string = (
            "[%(asctime)s.%(msecs)03d] [%(name)s] [%(levelname)s] %(message)s"
        )
    else:
        format_string = "%(message)s"

    colored_formatter = ColoredFormatter(format_string)
    colored_formatter.datefmt = "%Y-%m-%d %H:%M:%S"

    formatter = logging.Formatter(format_string)
    formatter.datefmt = "%Y-%m-%d %H:%M:%S"

    stream_handler = logging.StreamHandler()

    if os.environ.get("SYSTEM_DEBUG") == "true":
        stream_handler.setLevel(logging.DEBUG)
    else:
        stream_handler.setLevel(logging.INFO)

    stream_handler.setFormatter(colored_formatter)

    logger.addHandler(stream_handler)


def create_logger(identifier: str, with_decorators: bool = True) -> logging.Logger:
    """Create a logger for the stage (all other loggers will be children).

    :param stage_id: The ID of the stage logger
    :param with_decorators: If False only the message will be printed (no timestamp or level name)

    :returns: The stage logger
    """

    logger = logging.getLogger(identifier)

    _configure(logger, with_decorators)

    return logger



"""
@File       xcwarnings.py
@Brief      Module that detects new build warnings
@Author     rajaber
@Date       03-22-2021
@copyright  Microsoft Corporation. All rights reserved.
"""
import argparse
from collections import defaultdict
import json
import logging
import os
import re
import sys
from typing import ClassVar, Pattern


from xcwarnings.logger import create_logger
from xcwarnings.build_warning import BuildWarning, LocationalBuildWarning
from xcwarnings.build_warning_diff_report import BuildWarningDiffReport


class XcwarningsChecker:
    """A checker that detects new build warnings."""

    FILE_SCOPE_WARNING_PATTERN: ClassVar[Pattern] = re.compile(
        r"^(.*):(\d+):(\d+): warning: (.*)$"
    )
    TARGET_SCOPE_WARNING_PATTERN: ClassVar[Pattern] = re.compile(
        r"^warning: (.*) \(in target '(.*)' from project '(.*)'\)$"
    )
    TYPE_CHECK_PATTERN: ClassVar[Pattern] = re.compile(
        r".* '.*' took [0-9]+ms to type-check \(limit: [0-9]+ms\)"
    )

    xcodebuild_log_path: str
    defined_issues_file_path: str
    source_root: str
    skip_type_checks: bool
    log: logging.Logger

    def __init__(
        self,
        xcodebuild_log_path: str,
        defined_issues_file_path: str,
        source_root: str,
        skip_type_checks: str = False,
        log: logging.Logger | None = None,
    ):
        self.xcodebuild_log_path = xcodebuild_log_path
        self.defined_issues_file_path = defined_issues_file_path
        self.source_root = os.path.normpath(source_root)
        self.skip_type_checks = skip_type_checks

        if log:
            self.log = log.getChild(__name__)
        else:
            self.log = create_logger(__name__)

    def read_known_issues(self) -> dict[BuildWarning, int]:
        """Reads known build issues per optional configuration file.

        :returns: Dictionary of known build issues where the keys are the warnings of type BuildWarning,
        and the values are the counts of how many such warnings are expected in their respective files
        or project targets.
        Note: Those BuildWarning instances are not expected to contain line and column numbers because
        those tend to change as new content is added/removed.
        """

        return self.read_issues(self.defined_issues_file_path)

    def read_issues(self, issues_path: str) -> dict[BuildWarning, int]:
        """Reads build issues per the specified configuration file.

        :returns: Dictionary of build issues where the keys are the warnings of type BuildWarning,
        and the values are the counts of how many such warnings are expected in their respective files
        or project targets.
        Note: Those BuildWarning instances are not expected to contain line and column numbers because
        those tend to change as new content is added/removed.
        """

        with open(issues_path, encoding="utf-8") as file:
            data_file = file.read()

        issues_list = json.loads(data_file)

        return {
            LocationalBuildWarning.from_dict(issue): issue["count"]
            for issue in issues_list
        }

    def scan_line_for_warning(self, line: str) -> BuildWarning | None:
        """Scans given line from build output for a warning.

        :param line: Contents of a line of a build output file (as a string)

        :returns: BuildWarning instance or None if no warning is found in this line.
        """

        file_scope_warning_search = XcwarningsChecker.FILE_SCOPE_WARNING_PATTERN.search(
            line
        )

        if file_scope_warning_search:
            file_path = os.path.normpath(file_scope_warning_search.group(1))
            line_number = int(file_scope_warning_search.group(2))
            column = int(file_scope_warning_search.group(3))
            warning_statement = file_scope_warning_search.group(4)

            if not (file_path + os.sep).startswith(
                os.path.abspath(self.source_root) + os.sep
            ):
                # Exclude files not under source root
                return None

            relative_file_path = os.path.relpath(file_path, self.source_root)
            build_warning = LocationalBuildWarning.create_locational_file_warning(
                relative_file_path, line_number, column, warning_statement
            )
            return build_warning

        target_scope_warning_search = (
            XcwarningsChecker.TARGET_SCOPE_WARNING_PATTERN.search(line)
        )
        if target_scope_warning_search:
            warning_statement = target_scope_warning_search.group(1)
            target = target_scope_warning_search.group(2)
            project = target_scope_warning_search.group(3)
            build_warning = BuildWarning.create_target_warning(
                target, project, warning_statement
            )
            return build_warning

        return None

    def scan_for_warnings(self) -> [BuildWarning]:
        """Scans the given file for all build warnings

        :returns: An array of BuildWarning instances found. Empty array if none is found.
        Note: there may be duplicate build warnings if output file contains multiple lines refering to
        the same warning messages. They will show up as duplicate entries in the returned array.
        """
        with open(self.xcodebuild_log_path, encoding="utf-8") as file:
            data_file = file.readlines()

        warnings: [BuildWarning] = []

        for line in data_file:
            warning = self.scan_line_for_warning(line)
            if not warning:
                continue

            if self.skip_type_checks and XcwarningsChecker.TYPE_CHECK_PATTERN.match(
                warning.warning_statement
            ):
                continue

            warnings.append(warning)

        return warnings

    def get_warnings(self) -> set[BuildWarning]:
        """Gets all warnings found in given output file, deduped, and aggregated on the file level.

        :returns: An array of BuildWarning instances found. Empty array if none is found.
        Note: Duplicate build warnings in the output file with multiple lines refering to
        the same warning messages are deduped and will count as one error.
        """

        self.log.info(
            f"Extracting warnings from build output file: {self.xcodebuild_log_path}"
        )

        warnings = self.scan_for_warnings()
        warnings_deduped = set(warnings)
        self.log.info(
            f"Total number of warning instances found in build log file: {len(warnings_deduped)}."
        )

        return warnings_deduped

    @staticmethod
    def get_aggregated_warnings(
        warnings: set[LocationalBuildWarning | BuildWarning],
    ) -> dict[BuildWarning, int]:
        """Aggregates warnings so that all warnings in one file or target, have one entry in the output dictionary
        along with a count of how many times it occured

        :param warnings: Array of warnings found

        :returns: A dictionary of BuildWarning instances found. Each entry corresponds to one file or project/target.
        """

        warning_dict: dict[BuildWarning, int] = defaultdict(int)
        for warning in warnings:
            if isinstance(warning, LocationalBuildWarning):
                warning_dict[warning.locationless_copy()] += 1
            else:
                warning_dict[warning] += 1

        return warning_dict

    def generate_baseline(self) -> None:
        """Gets baseline for all known issues identified in the output file path, and writes them
        to a known issues file at the given path.

        This writes known issues in JSON format at output_known_issues_file.
        """

        self.log.info("Generating baseline 📈")
        warnings = self.get_warnings()
        warnings_with_count = XcwarningsChecker.get_aggregated_warnings(warnings)
        supressions = []

        for unique_warning, count in warnings_with_count.items():
            supression_entry = unique_warning.to_dict()
            supression_entry["count"] = count
            supressions.append(supression_entry)

        supressions.sort(
            key=lambda s: (
                s.get("project", ""),
                s.get("target", ""),
                s.get("file_path", ""),
                s.get("warning", ""),
            )
        )

        known_issues_json = json.dumps(supressions, indent=4)

        with open(self.defined_issues_file_path, "w", encoding="utf-8") as text_file:
            text_file.write(known_issues_json)

        self.log.info(f"Baseline generated at file: {self.defined_issues_file_path}")

    def analyze(
        self,
    ) -> BuildWarningDiffReport:
        """Compares warnings present in the given output file with known warnings in
        defined_issues_file_path.

        :returns: An instance of BuildWarningDiffReport
        """
        known_issues = self.read_known_issues()
        current_warnings = self.get_warnings()
        aggregated_warnings = XcwarningsChecker.get_aggregated_warnings(
            current_warnings
        )
        return BuildWarningDiffReport.calculate_diff_report(
            known_issues, aggregated_warnings, current_warnings, self.log
        )


def handle_arguments() -> int:
    """Parses xcode build output file, and confirms there are no new warning errors.
    When --generate_baseline argument is passed, a baseline is generated with warnings found
    in given output log.

    :returns: An exit code signifying success or failure of the build warnings gate.
    """

    log = create_logger(__name__)

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--xcodebuild-log-path",
        dest="xcodebuild_log_path",
        help="Path to the xcodebuild output file",
        required=True,
    )
    parser.add_argument(
        "--defined-issues-file-path",
        dest="defined_issues_file_path",
        required=True,
        help="Full path to a JSON file with known build warnings",
    )
    parser.add_argument(
        "--source-root",
        dest="source_root",
        required=True,
        help="File path for the root of the source code",
    )
    parser.add_argument(
        "--generate-baseline",
        action="store_true",
        default=False,
        help="Whether a new baseline of known issues should be generated.",
    )

    args = parser.parse_args()

    log.info("Staring Build Warnings Detector ⚠️ ")

    checker = XcwarningsChecker(
        args.xcodebuild_log_path,
        args.defined_issues_file_path,
        args.source_root,
        log,
    )

    if args.generate_baseline:
        checker.generate_baseline()
        return 0

    log.info("Analyzing current warnings against known issues if any🔬.")
    diff_report = checker.analyze()
    log.info("Build Warnings Gate finished running. 🔥")
    return 0 if diff_report.success() else 1


if __name__ == "__main__":
    sys.exit(handle_arguments())



"""
@File       build_warning.py
@Brief      An Xcode build warning
@Author     rajaber
@Date       03-22-2021
@copyright  Microsoft Corporation. All rights reserved.
"""

from typing import Union

from xcwarnings.build_warning_scope import BuildWarningScope


class BuildWarning:
    # pylint: disable = too-many-arguments
    """BuildWarning contains metadata about a warning in an xcode build output file"""

    def __init__(
        self,
        file_path: str | None,
        target: str | None,
        project: str | None,
        warning_statement: str,
    ) -> None:
        self.file_path = file_path
        self.target = target
        self.project = project
        self.warning_statement = warning_statement
        if self.file_path is not None:
            self.scope = BuildWarningScope.FILE
        else:
            self.scope = BuildWarningScope.TARGET

    # pylint: enable = too-many-arguments

    def __repr__(self) -> str:
        if self.scope == BuildWarningScope.FILE:
            return (
                f"BuildWarning(Warning in file: {self.file_path}. "
                + f"Warning: {self.warning_statement})"
            )

        if self.scope == BuildWarningScope.TARGET:
            return (
                f"BuildWarning(Warning in target: {self.target}. "
                + f"Project: {self.project}. Warning: {self.warning_statement})"
            )

        raise ValueError(f"Unexpected scope: {self.scope}")

    def __eq__(self, other: object):
        """Checks if currrent instance is eual to the other instance

        :param other: object to compare to current instance for equality

        :returns: True if objects are equal; otherwise, false.
        """
        if not isinstance(other, BuildWarning):
            return False

        return (
            (self.file_path == other.file_path)
            and (self.warning_statement == other.warning_statement)
            and (self.project == other.project)
            and (self.target == other.target)
        )

    def __ne__(self, other: object) -> bool:
        """Checks if currrent instance is not equal to the other instance

        :param other: object to compare to current instance for equality

        :returns: True if objects are not equal; otherwise, true.
        """
        return not self.__eq__(other)

    def __hash__(self) -> int:
        """Returns hash of current instance

        :returns: Hash of BuildWarning instance
        """
        return hash(self.__repr__())

    def to_dict(self) -> dict[str, object]:
        """Serializes BuildWarning instance into a dictionary."""
        warning_dict: dict[str, object] = {}
        warning_dict["warning"] = self.warning_statement
        if self.scope == BuildWarningScope.FILE:
            warning_dict["file_path"] = self.file_path
        elif self.scope == BuildWarningScope.TARGET:
            warning_dict["target"] = self.target
            warning_dict["project"] = self.project
        else:
            raise ValueError(f"Unexpected scope: {self.scope}")
        return warning_dict

    @staticmethod
    def create_file_warning(
        file_path: str,
        warning_statement: str,
    ) -> "BuildWarning":
        """Creates a file level warning

        :param file_path: Path to the file where the warning occurs, relative to the repo root.
        :param warning_statement: Warning statement

        :returns: A file level warning
        """
        return BuildWarning(file_path, None, None, warning_statement)

    @staticmethod
    def create_target_warning(
        target: str, project: str, warning_statement: str
    ) -> "BuildWarning":
        """Creates a target level warning

        :param target: Target where the warning occurs
        :param project: Project where the warning occurs
        :param warning_statement: Warning statement

        :returns: A target level warning
        """
        return BuildWarning(None, target, project, warning_statement)

    @staticmethod
    def from_dict(input_dict: dict[str, object]) -> "BuildWarning":
        """Creates an instance of BuildWarning from given dictinoary.

        :param input_dict: dictionary serialization of BuildWarning

        :returns: An instance of build warning
        """
        return BuildWarning(
            input_dict.get("file_path"),
            input_dict.get("target"),
            input_dict.get("project"),
            input_dict["warning"],
        )


class LocationalBuildWarning(BuildWarning):
    # pylint: disable = too-many-arguments
    """BuildWarning contains metadata about a warning in an xcode build output file"""

    def __init__(
        self,
        file_path: str | None,
        line_number: str,
        column: str,
        project: str | None,
        warning_statement: str,
    ) -> None:
        super().__init__(file_path, None, project, warning_statement)
        assert line_number is not None
        assert column is not None
        self.line_number = line_number
        self.column = column

    def locationless_copy(self) -> BuildWarning:
        """Return a version of self without the location information set."""
        return BuildWarning.from_dict(self.to_dict())

    # pylint: enable = too-many-arguments

    def __repr__(self) -> str:
        if self.scope == BuildWarningScope.FILE:
            return (
                f"LocationalBuildWarning(Warning in file: {self.file_path}:{self.line_number}:{self.column}. "
                + f"Warning: {self.warning_statement})"
            )

        if self.scope == BuildWarningScope.TARGET:
            return (
                f"LocationalBuildWarning(Warning in target: {self.target}. "
                + f"Project: {self.project}. Warning: {self.warning_statement})"
            )

        raise ValueError(f"Unexpected scope: {self.scope}")

    def __eq__(self, other: object):
        """Checks if currrent instance is eual to the other instance

        :param other: object to compare to current instance for equality

        :returns: True if objects are equal; otherwise, false.
        """
        if not isinstance(other, LocationalBuildWarning):
            return False

        if not super().__eq__(other):
            return False

        return (self.line_number == other.line_number) and (self.column == other.column)

    def __ne__(self, other: object) -> bool:
        """Checks if currrent instance is not equal to the other instance

        :param other: object to compare to current instance for equality

        :returns: True if objects are not equal; otherwise, true.
        """
        return not self.__eq__(other)

    def __hash__(self) -> int:
        """Returns hash of current instance

        :returns: Hash of BuildWarning instance
        """
        return hash(self.__repr__())

    def to_dict(self) -> dict[str, object]:
        """Serializes BuildWarning instance into a dictionary."""
        warning_dict = super().to_dict()
        if self.scope == BuildWarningScope.FILE:
            warning_dict["line_number"] = self.line_number
            warning_dict["column"] = self.column
        return warning_dict

    @staticmethod
    def create_locational_file_warning(
        file_path: str,
        line_number: int | None,
        column: int | None,
        warning_statement: str,
    ) -> "LocationalBuildWarning":
        """Creates a file level warning

        :param file_path: Path to the file where the warning occurs, relative to the repo root.
        :param line_number: Line number where the warning occurs
        :param column: Column within the line where the warning occurs
        :param warning_statement: Warning statement

        :returns: A file level warning
        """
        return LocationalBuildWarning(
            file_path, line_number, column, None, warning_statement
        )

    @staticmethod
    def from_dict(
        input_dict: dict[str, object]
    ) -> Union["LocationalBuildWarning", BuildWarning]:
        """Creates an instance of BuildWarning from given dictinoary.

        :param input_dict: dictionary serialization of BuildWarning

        :returns: An instance of build warning
        """

        if "line_number" in input_dict and "column" in input_dict:
            return LocationalBuildWarning(
                input_dict.get("file_path"),
                input_dict.get("line_number"),
                input_dict.get("column"),
                input_dict.get("project"),
                input_dict["warning"],
            )

        return BuildWarning.from_dict(input_dict)
